local name, private = ...
local astrolabe = DongleStub("Astrolabe-1.0")
local dbversion = 3

local queue = private.objects.queue
local list = private.objects.list
local dict = private.objects.dict

-- This part of Excavatinator uses some of the base data, but is otherwise entirely separate from the rest.

local currencyNames = {}
for k, v in pairs(private.currencies) do
    local name = GetCurrencyInfo(k)
    currencyNames[name] = v
end

local tracker = private.objects.base()

local transfers = {
    get = dict(),
    send = dict(),
}

local TRANSFERS_PER_SECOND = 50
local TRANSFER_INTERVAL = 1

------------------------------------------------------------
----------------------- Icon tracker -----------------------
------------------------------------------------------------

tracker.icons = queue()
tracker.icons.inUse = queue()

globaltracker = tracker -- TODO: Remove!

function tracker.icons:getIcon()
    local r = self:dequeue()
    if not r then
        --print("Creating a new icon")
        r = CreateFrame("Button", nil, UIParent)
        r:SetSize(16, 16)
        r:SetNormalTexture("Interface/Archeology/Arch-Icon-Marker")
    else
        --print("Using an existing icon")
    end
    self.inUse:enqueue(r)
    return r
end

function tracker.icons:storeIcon(icon)
    if self.inUse:remove(icon) then
        --print("Returning icon")
        self:enqueue(icon)
    end
end

function tracker.icons:storeAll()
    for i = 1, #self.inUse do
        local icon = self.inUse:dequeue()
        self:enqueue(icon)
    end
end

------------------------------------------------------------
--------------------- Minimap display ----------------------
------------------------------------------------------------

local interval = 1

tracker.passed = 0
tracker.__eventFrame:SetScript("OnUpdate", function(self, elapsed)
    --print(elapsed)
    self.__parent.passed = self.__parent.passed + elapsed
    if self.__parent.passed > interval then
        self.__parent.passed = 0
        self.__parent:updateMinimap()
    end
end)

-- The tracker needs to check that a gathering spot has an icon.
-- If there is no icon, it needs to create one. However, not
-- having an icon does not mean there is no table in its place
-- (because of how WoW saves the icon field). The field can still
-- be nil, so placing an empty table there in that case ensures
-- that there will be no error when checking the length of the
-- table. And having one single table that gets used, saves memory.
local emptyTable = {}

function tracker:updateMinimap()
    -- Shows icons within a 250 yard radius.
    -- Hides all other icons.
    -- Copied almost completely from 1.6.1.
    
    local db = ExcavatinatorNodeDB
    if db.show then
        local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
        if db[c] then
            if db[c][z] then
                for i=1, #db[c][z] do
                    local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
                    if distance < 250 then
                        if not db[c][z][i].icon then
                            --print("No icon")
                            db[c][z][i].icon = emptyTable
                        end
                        --if #db[c][z][i].icon == 0 then
                        if not db[c][z][i].icon[0] then
                            --print("Zero-length icon")
                            local icon = self.icons:getIcon()
                            globalicon = icon
                            astrolabe:PlaceIconOnMinimap(icon, c, z, db[c][z][i].x, db[c][z][i].y)
                            db[c][z][i].icon = icon
                        end
                    else
                        self.icons:storeIcon(db[c][z][i].icon)
                        astrolabe:RemoveIconFromMinimap(db[c][z][i].icon)
                        db[c][z][i].icon = nil
                    end
                end
            end
        end
    end
end

function tracker:addPosition(currency, optc, optz, optx, opty)
    -- Adds a position to the node database.
    -- If no position is given, adds the player's current position.
    -- If one is given, adds that.
    -- Copied almost completely from 1.6.1.
    
    -- Do not add the position if Astrolabe does not know the map!
    --print("here 1")
    --if not astrolabe:GetMapID() then return end
    --print("here 2")
    
    local db = ExcavatinatorNodeDB
    if db.track then
        -- Add or update the position
        local c, z, x, y
        if (optc and optz and optx and opty) then
            c, z, x, y = optc, optz, optx, opty
        else
            c, z, x, y = astrolabe:GetCurrentPlayerPosition()
        end
        if not db[c] then db[c] = {} end -- If no entry for the current continent, create one
        if not db[c][z] then db[c][z] = {} end -- If no entry for the current zone, create one
        local found = 0
        for i=1, #db[c][z] do
            local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
            --print("Distance", distance)
            if distance < 10 then
                --print("Found an old position!")
                -- It's an old position; add one artifact found
                db[c][z][i].gathers = db[c][z][i].gathers + 1
                if currency > 0 then db[c][z][i].race = currency end -- In case of importing, don't overwrite the race
                found = i
                break
            end
        end
        if found == 0 then
            --print("Found a new position!")
            -- It's a new position; add it to the database and make an icon for the position (if it's in the zone the player is actually in)
            db[c][z][#db[c][z] + 1] = { x = x, y = y, gathers = 1, race = currency }
            local cr, zr = astrolabe:GetCurrentPlayerPosition()
            if c == cr and z == zr then
                local icon = self.icons:getIcon()
                astrolabe:PlaceIconOnMinimap(icon, c, z, x, y)
                db[c][z][#db[c][z]].icon = icon
            end
        end
    end
end

tracker:RegisterEvent("ZONE_CHANGED_NEW_AREA")
function tracker:ZONE_CHANGED_NEW_AREA()
    self.icons:storeAll()
    self:updateMinimap()
end

tracker:RegisterEvent("CHAT_MSG_CURRENCY")
function tracker:CHAT_MSG_CURRENCY(msg)
    local raceID = 0
    for k, v in pairs(currencyNames) do
        if msg:find(k) then
            raceID = v
        end
    end
    --print("Found fragments for race " .. raceID)
    if raceID == 0 then return end
    self:addPosition(raceID)
end

------------------------------------------------------------
-------------------------- Popups --------------------------
------------------------------------------------------------

local popups = dict()

local popupList = {
    ["EXCAVATINATOR_GET_OFFER"] = {
        text = "%s is offering to share their node database. Accept?",
        button1 = "Yes",
        button2 = "No",
        OnAccept = function(self, playerName)
            local transfer = transfers.get:get(playerName)
            transfer.mode = "get"
            transfer.state = "getting"
            transfer:send_ACCEPT_OFFER()
        end,
        OnCancel = function(self, playerName)
            local transfer = transfers.get:get(playerName)
            transfer:send_DENY()
        end,
        timeout = 30,
        whileDead = true,
        hideOnEscape = true,
    },
    ["EXCAVATINATOR_GET_REQUEST"] = {
        text = "%s is requesting that you share your node database with them. Accept?",
        button1 = "Yes",
        button2 = "No",
        OnAccept = function(self, playerName)
            local transfer = transfers.send:get(playerName)
            transfer.mode = "send"
            transfer:prepareData()
            transfer:start()
        end,
        OnCancel = function(self, playerName)
            local transfer = transfers.send:get(playerName)
            transfer:send_DENY()
        end,
        timeout = 30,
        whileDead = true,
        hideOnEscape = true,
    },
    --["EXCAVATINATOR_DELETEDB"] = {
    --    text = "Are you sure you want to clear your node database? This can NOT be undone.",
    --    button1 = "Yes",
    --    button2 = "No",
    --    OnAccept = resetDB,
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_IMPORT_ARCHY"] = {
    --    text = "Import node data from Archy?",
    --    button1 = "Yes",
    --    button2 = "No",
    --    OnAccept = importFromArchy,
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},

    --["EXCAVATINATOR_YOU_ARE_BUSY"] = {
    --    text = "You are busy.",
    --    button1 = "OK",
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_SEND_REQUEST"] = {
    --    text = "Send a database transfer request to %s?",
    --    button1 = "Yes",
    --    button2 = "No",
    --    OnAccept = function() sendRequest(currentTransfer.player) end,
    --    OnCancel = cancel,
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_SEND_OFFER"] = {
    --    text = "Offer to share your database with %s?",
    --    button1 = "Yes",
    --    button2 = "No",
    --    OnAccept = function() sendOffer(currentTransfer.player) end,
    --    OnCancel = cancel,
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_TRANSFER_DENIED"] = {
    --    text = "%s has denied your request.",
    --    button1 = "OK",
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_TRANSFER_BUSY"] = {
    --    text = "%s is busy.",
    --    button1 = "OK",
    --    timeout = 0,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
    --["EXCAVATINATOR_TRANSFER_TIMEOUT"] = {
    --    text = "Your transfer has timed out.",
    --    button1 = "OK",
    --    timeout = TIMEOUT_DIALOG,
    --    whileDead = true,
    --    hideOnEscape = true,
    --},
}

for k, v in pairs(popupList) do
    StaticPopupDialogs[k] = v
end

popups:merge(popupList)

function popups:show(key, ...)
    -- Formats the text in the popup using the variable arguments as arguments,
    -- then displays the popup.
    local text = StaticPopupDialogs[key].text
    StaticPopupDialogs[key].text = format(text, ...)
    local popup = StaticPopup_Show(key)
    StaticPopupDialogs[key].text = text
    return popup
end

--function test()
--    local popup = popups:show("TEST")
--    popup.data = "one"
--    popup.data2 = "two"
--end

------------------------------------------------------------
----------------------- Node sharing -----------------------
------------------------------------------------------------

--[[

Successful offer v1:
You: OFFER
Oth: ACCEPT_OFFER
You: DATASTART:n:3
You: Repeat n times:
    DATANODE:zone:floor:x:y:race
You: END

Successful offer v2:
You: OFFER
You: VERSION2
Oth: ACCEPT_OFFER
You: DATASTART:n:3
You: Repeat n times:
    D:n:<data string>*
You: END

* Data string: node;node;node...

]]

RegisterAddonMessagePrefix("Excavatinator_TR")

local transfer = private.objects.base()

function transfer:init(target)
    self:RegisterEvent("CHAT_MSG_ADDON")
    
    -- This version number is the version of the transfer system.
    -- Version 1 is the one used in 1.6.1 and earlier.
    -- Version 2 is the one used in version 2.
    self.version = 1
    self.state = "none" -- "none", "getting" or "sending"
    self.target = target
    self.mode = "" -- "get" or "send"
    self.data = list() -- For use with v1 clients, or turning into v2 data.
    self.datastring = "" -- For use with v2 clients
    self.progress = 0
    self.max = 0
    self.stepValue = 1
    self.ended = false
end

function transfer:stop()
    -- After the transfer has ended, stop this transfer object from
    -- updating itself.
    if self.updateFrame then
        self.updateFrame:SetScript("OnUpdate", nil)
    end
    self:UnregisterAllEvents()
end

function transfer:prepareData()
    -- Turn the node list into a list of strings to send
    local db = ExcavatinatorNodeDB
    for zoneID, zoneTbl in pairs(db) do
        if type(zoneID) == "number" then
            for floor, floorTbl in pairs(zoneTbl) do
                for num, node in pairs(floorTbl) do
                    self.data:add(format("DATANODE:%d:%d:%g:%g:%d", zoneID, floor, node.x, node.y, node.race))
                end
            end
        end
    end
    
    -- If this is a v2 transfer, turn the data list into a single data string to send.
    if self.version == 1 then return end
    self.datastring = ""
    for i = 1, #self.data do
        self.datastring = self.datastring .. self.data[i]:sub(9) .. ";"
    end
    self.datastring = self.datastring:sub(1, -2) -- Cuts off the trailing ";"
end

function transfer.onUpdate(self, elapsed)
    -- self is the update frame here
    self.passed = self.passed + elapsed
    if self.passed > TRANSFER_INTERVAL then
        self.__parent:step()
        self.passed = 0
    end
end

function transfer:start()
    if self.state ~= "sending" then return end
    if self.mode ~= "send" then return end
    -- Starts sending data to the target player.
    local length = self.datastring:len()
    self.max = Math.ceil(length / 250)
    SendAddonMessage(format("DATASTART:%d:3", self.max))
    
    self.updateFrame = CreateFrame("Frame")
    self.updateFrame.__parent = self
    self.updateFrame.passed = 0
    self.updateFrame:SetScript("OnUpdate", self.onUpdate)
end

function transfer:step()
    if self.state ~= "sending" then return end
    -- Sends the next piece of data to the other player.
    if self.ended then
        self:send_END()
        self:stop()
        return
    end
    if self.version == 1 then
        -- Send as many data strings as we can
        for i = self.stepValue, self.stepValue + TRANSFERS_PER_SECOND - 1 do
            SendAddonMessage("Excavatinator_TR", self.data[i], "WHISPER", self.target)
            self.progress = self.progress + 1
            if i == #self.data then
                -- There is no more data to send
                self.ended = true
                break
            end
        end
    end
    if self.version == 2 then
        for i = self.stepValue, self.stepValue + TRANSFERS_PER_SECOND - 1 do
            local message = self.datastring:sub((i - 1) * 250 + 1, i * 250)
            SendAddonMessage("Excavatinator_TR", "D:" .. i .. ":" .. message, "WHISPER", self.target)
            self.progress = self.progress + 1
            if i * 250 >= self.datastring:len() then
                -- There is no more data to send
                self.ended = true
                break
            end
        end
    end
    
    self.stepValue = self.stepValue + TRANSFERS_PER_SECOND
end

function transfer:parse()
    -- Parses the data received from the other player
    if self.version == 1 then
        for i = 1, #self.data do
            local zone, floor, x, y, race = strsplit(":", self.data[i])
            tracker:addPosition(tonumber(race), tonumber(zone), tonumber(floor), tonumber(x), tonumber(y))
        end
    end
    if self.version == 2 then
        if self.progress ~= self.max then
            self:err_MISSING_DATA()
            self:stop()
            return
        end
        local nodeString = table.concat(self.data)
        local nodeList = { strsplit(";", nodeString) }
        for num, nodeString in pairs(nodeList) do
            local zone, floor, x, y, race = strsplit(":", nodeString)
            tracker:addPosition(tonumber(race), tonumber(zone), tonumber(floor), tonumber(x), tonumber(y))
        end
    end
end

function transfer:err_BUSY()
    -- Error - triggered when this transfer cannot be initiated,
    -- because a transfer is already running for this player.
    self:stop()
end

function transfer:err_MISSING_DATA()
    -- Called if, when the END signal is received, there is missing data.
    self:stop()
end

function transfer:get_VERSION(data)
    -- If the other player is running a client at v2 or higher,
    -- this message will be sent immediately on receiving an OFFER
    -- or REQUEST message.
    -- In v2, this is only the version number. For future
    -- proofing, only the part up to a ":" will be read.
    local version = tonumber(strsplit(":", data))
    self.version = version > 2 and 2 or version
end

function transfer:get_OFFER()
    -- Got an offer from another player.
    -- We do not yet know whether the other side is a v1 or v2 client.
    if transfers.get:get(self.target) then self:err_BUSY() return end -- Only one transfer each way per player at a time.
    transfers.get:set(self.target, self)
    self.mode = "get"
    
    -- Show the popup
    local popup = popups:show("EXCAVATINATOR_GET_OFFER")
    popup.data = self.target
    
    -- Send the version signal
    self:send_VERSION()
end

function transfer:get_REQUEST()
    -- Got a request from another player.
    -- We do not yet know whether the other side is a v1 or v2 client.
    if transfers.send:get(self.target) then self:err_BUSY() return end -- Only one transfer each way per player at a time.
    transfers.send:set(self.target, self)
    self.mode = "send"
    
    -- Show the popup
    local popup = popups:show("EXCAVATINATOR_GET_REQUEST")
    popup.data = self.target
    
    -- Send the version signal
    self:send_VERSION()
end

function transfer:get_ACCEPT_OFFER()
    -- The other player has accepted our offer to send data.
    -- Start sending data
    if self.mode ~= "send" then self:stop() return end
    self.state = "sending"
    self:start()
end

function transfer:get_DATASTART(data)
    -- The other player has started sending data.
    self.max = tonumber(strsplit(":", data))
    self.progress = 1
    self.data:clear()
end

function transfer:get_DATANODE(data)
    -- The other player has sent a v1 data string.
    if self.version ~= 1 then self:stop() return end
    if self.mode ~= "get" then self:stop() return end
    if self.state ~= "getting" then self:stop() return end
    self.data:add(data)
end

function transfer:get_D(data)
    -- The other player has sent a v2 data chunk.
    if self.version ~= 2 then self:stop() return end
    if self.mode ~= "get" then self:stop() return end
    if self.state ~= "getting" then self:stop() return end
    local num, string = strsplit(":", data)
    self.progress = self.progress + 1
    self.data[num] = string
    --D:n:<data string>*
    --self.datastring = self.datastring .. data
end

function transfer:get_END()
    -- The other player has sent the END signal.
    -- The data transfer is complete; parse the data and store it.
    self.mode = "none"
    self.state = "none"
    self:parse()
    self:stop()
end

function transfer:get_CANCEL()
    -- The other player has cancelled the transfer.
    -- This was added in version 2; it does not have a v1 equivalent.
    self:stop()
end

function transfer:get_BUSY()
    -- BUSY is a code that means the addon is busy, and can't
    -- initiate any more transfers. In the case of v1, this
    -- means the target player is running any transfer at all;
    -- in the case of v2, it simply means that a transfer
    -- between these two players is already running.
    -- This transfer should cancel in that case.
    self:stop()
end

function transfer:get_DENY()
    -- The target player has denied the transfer.
    -- This transfer has been cancelled.
    self:stop()
end

function transfer:send_VERSION()
    -- Sends this addon's transfer version number to the other player.
    SendAddonMessage("Excavatinator_TR", "VERSION2", "WHISPER", self.target)
end

function transfer:send_OFFER()
    -- Sends a data offer to the target player.
    SendAddonMessage("Excavatinator_TR", "OFFER", "WHISPER", self.target)
end

function transfer:send_REQUEST()
    -- Sends a data offer to the target player.
    SendAddonMessage("Excavatinator_TR", "REQUEST", "WHISPER", self.target)
end

function transfer:send_ACCEPT_OFFER()
    -- The player has accepted the database; respond with ACCEPT_OFFER.
    SendAddonMessage("Excavatinator_TR", "ACCEPT_OFFER", "WHISPER", self.target)
end

function transfer:send_DATA()
    -- Dummy function. The data methods handle this. Remove later.
end

function transfer:send_END()
    -- Sends the END signal to tell the other client that the transfer is over.
    SendAddonMessage("Excavatinator_TR", "END", "WHISPER", self.target)
    print("Transfer ended.")
    self:stop()
end

function transfer:send_CANCEL()
    -- The player has cancelled the transfer; tell the other client.
    SendAddonMessage("Excavatinator_TR", "CANCEL", "WHISPER", self.target)
    self:stop()
end

function transfer:send_BUSY()
    -- Sends the BUSY signal to the other player.
    SendAddonMessage("Excavatinator_TR", "BUSY", "WHISPER", self.target)
    self:stop()
end

function transfer:send_DENY()
    -- Sends the DENY signal to the other player, cancelling the transfer.
    SendAddonMessage("Excavatinator_TR", "DENY", "WHISPER", self.target)
    self:stop()
end

function transfer:CHAT_MSG_ADDON(prefix, msg, mode, sender)
    if prefix ~= "Excavatinator_TR" then return
    elseif sender ~= self.target then return -- This isn't for this transfer.
    elseif msg:sub(1, 9) == "BROADCAST" then return -- The tracker handles broadcasts.
    elseif msg:sub(1, 4) == "BUSY" then self:get_BUSY() -- The addon on the other end is busy.
    elseif msg:sub(1, 7) == "VERSION" then -- The other client is a v2 client (or higher), and has sent version data.
        self:get_VERSION(msg:sub(8))
    elseif msg:sub(1, 9) == "DATASTART" then -- The other client has started sending data.
        self:get_DATASTART(msg:sub(11))
    elseif msg:sub(1, 4) == "DATA" then -- The other client has sent a v1 data string
        self:get_DATA(msg:sub(6))
    elseif msg:sub(1, 1) == "D" then -- The other client has sent a v2 data chunk
        self:get_D(msg:sub(3))
    end
end

-- This function reacts to any messages that initiate data
-- transfers, and also broadcasts.
tracker:RegisterEvent("CHAT_MSG_ADDON")
function tracker:CHAT_MSG_ADDON(prefix, msg, mode, sender)
    if prefix ~= "Excavatinator_TR" then return end
    if msg:sub(1, 9) == "BROADCAST" then -- It's a broadcast of a single node gather
        local db = ExcavatinatorNodeDB
        -- It's a single node broadcast
        if not db.broadcast[mode] then return end -- For GUILD or unknown modes
        if type(db.broadcast[mode] == "table") then
            if not db.broadcast[mode][sender] then return end
        end
        local c, z, x, y, race = strsplit(":", string.sub(msg, 10))
        self:addPosition(tonumber(race), tonumber(c), tonumber(z), tonumber(x), tonumber(y))
    end
    if mode ~= "WHISPER" then return end -- Data transfers allowed only through whispers
    if msg:sub(1, 5) == "OFFER" then -- Data offer
        local t = transfer()
        t:init(sender)
        t:get_OFFER()
    end
    if msg:sub(1, 7) == "REQUEST" then -- Data request
        local t = transfer()
        t:init(sender)
        t:get_REQUEST()
    end
end

------------------------------------------------------------
--------------------------- Load ---------------------------
------------------------------------------------------------

local defaultDB = {
    broadcast = {
        GUILD = true,
        WHISPER = {},
        CHANNEL = {},
    },
    version = dbversion,
    show = true,
    track = true,
}

function tracker:VerifyDB()
    -- First, make sure the DB is a table
    if type(ExcavatinatorNodeDB) ~= "table" then
        ExcavatinatorNodeDB = defaultDB
    end
    
    -- If version is below 3, lose the settings. Any data still there is irrelevant now anyway.
    if ExcavatinatorNodeDB.version < 3 then
        ExcavatinatorNodeDB.broadcast = defaultDB.broadcast
        ExcavatinatorNodeDB.version = 3
        ExcavatinatorNodeDB.show = true
        ExcavatinatorNodeDB.track = true
        print(L("Your node database settings were out of date, and have been reset. All node data is intact."))
    end
    
    -- Make sure the broadcast list is properly formatted
    if type(ExcavatinatorNodeDB.broadcast) ~= "table" then
        ExcavatinatorNodeDB.broadcast = defaultDB.broadcast
    end
    if type(ExcavatinatorNodeDB.broadcast.WHISPER) ~= "table" then
        ExcavatinatorNodeDB.broadcast.WHISPER = {}
    end
    if type(ExcavatinatorNodeDB.broadcast.CHANNEL) ~= "table" then
        ExcavatinatorNodeDB.broadcast.CHANNEL = {}
    end
    if type(ExcavatinatorNodeDB.broadcast.GUILD) ~= "boolean" then
        ExcavatinatorNodeDB.broadcast.GUILD = true
    end
end

tracker:RegisterEvent("ADDON_LOADED")
function tracker:ADDON_LOADED(addon)
    if addon == name then
        self:VerifyDB()
        self:updateMinimap()
    end
end